export const CONTEXT_KEY_CHAT_MESSAGE_EDIT = 'chat-message-edit';
export const CONTEXT_KEY_CHAT_MESSAGE_ACTIONS = 'chat-message-actions';
export const CONTEXT_KEY_CHAT_FORM_ACTIONS = 'chat-form-actions';
